@@ -131,27 +131,34 @@ class Element(html5lib.treebuilders._base.Node):
             old_element = self.element.contents[-1]
             new_element = self.soup.new_string(old_element + node.element)
             old_element.replace_with(new_element)
+            self.soup._most_recent_element = new_element
         else:
-            self.element.append(node.element)
-            node.parent = self
+            self.soup.object_was_parsed(node.element, parent=self.element)
 
     def getAttributes(self):
         return AttrList(self.element)
 
     def setAttributes(self, attributes):
-        if attributes is not None and attributes != {}:
+        if attributes is not None and len(attributes) > 0:
+
+            converted_attributes = []
             for name, value in list(attributes.items()):
                 if isinstance(name, tuple):
-                    name = NamespacedAttribute(*name)
-                self.element[name] =  value
+                    new_name = NamespacedAttribute(*name)
+                    del attributes[name]
+                    attributes[new_name] = value
+
+            self.soup.builder._replace_cdata_list_attribute_values(
+                self.name, attributes)
+            for name, value in attributes.items():
+                self.element[name] = value
+
             # The attributes may contain variables that need substitution.
             # Call set_up_substitutions manually.
             #
             # The Tag constructor called this method when the Tag was created,
             # but we just set/changed the attributes, so call it again.
-            self.element.contains_substitutions = (
-                self.soup.builder.set_up_substitutions(
-                    self.element))
+            self.soup.builder.set_up_substitutions(self.element)
     attributes = property(getAttributes, setAttributes)
 
     def insertText(self, data, insertBefore=None):
